# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PyTorchTaskRunner module."""

import logging
import os
from copy import deepcopy
from typing import Iterator, Tuple

import numpy as np
import torch
import torch.nn as nn
import tqdm

from openfl.federated.task.runner import TaskRunner
from openfl.utilities import Metric, TensorKey, change_tags
from openfl.utilities.split import split_tensor_dict_for_holdouts

logger = logging.getLogger(__name__)


class PyTorchTaskRunner(nn.Module, TaskRunner):
    """PyTorch Model class for Federated Learning.

    Attributes:
        device (str): Compute device (default="cpu")
        required_tensorkeys_for_function (dict): A map of all the required
            tensors for each of the public functions in PyTorchTaskRunner.
        optimizer (Optimizer): The optimizer for the model.
        loss_fn (function): The loss function for the model.
        training_round_completed (bool): Flag to check if training round is
            completed.
        tensor_dict_split_fn_kwargs (dict): Arguments for the tensor
            dictionary split function.
    """

    def __init__(self, device: str = None, loss_fn=None, optimizer=None, **kwargs):
        """Initializes the PyTorchTaskRunner object.

        Args:
            device (str): Compute device (default="cpu").
            loss_fn (function): The loss function for the model.
            optimizer (Optimizer): The optimizer for the model.
            **kwargs: Additional parameters to pass to the functions.
        """
        super().__init__()
        TaskRunner.__init__(self, **kwargs)
        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        elif isinstance(device, str):
            self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            raise ValueError("Device must be None, a string, or a torch.device object.")

        # This is a map of all the required tensors for each of the public
        # functions in PyTorchTaskRunner
        self.required_tensorkeys_for_function = {}

        self.optimizer = optimizer
        self.loss_fn = loss_fn
        self.training_round_completed = False

        # overwrite attribute to account for one optimizer param (in every
        # child model that does not overwrite get and set tensordict) that is
        # not a numpy array
        self.tensor_dict_split_fn_kwargs.update({"holdout_tensor_names": ["__opt_state_needed"]})

    def rebuild_model(self, round_num, input_tensor_dict, validation=False):
        """Parse tensor names and update weights of model. Handles the
        optimizer treatment.

        Args:
            round_num (int): The current round number
            input_tensor_dict (dict): The input tensor dictionary
            validation (bool): Flag to check if it's validation

        Returns:
            None
        """
        if self.opt_treatment == "RESET":
            self.reset_opt_vars()
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)
        elif (
            self.training_round_completed
            and self.opt_treatment == "CONTINUE_GLOBAL"
            and not validation
        ):
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=True)
        else:
            self.set_tensor_dict(input_tensor_dict, with_opt_vars=False)

    def validate_task(self, col_name, round_num, input_tensor_dict, use_tqdm=False, **kwargs):
        """Validate Task.

        Run validation of the model on the local data.

        Args:
            col_name (str): Name of the collaborator.
            round_num (int): What round is it.
            input_tensor_dict (dict): Required input tensors (for model).
            use_tqdm (bool): Use tqdm to print a progress bar (Default=True).
            **kwargs: Additional parameters.

        Returns:
            global_output_dict (dict):  Tensors to send back to the aggregator.
            local_output_dict (dict):   Tensors to maintain in the local
                TensorDB.
        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        self.eval()
        self.to(self.device)

        loader = self.data_loader.get_valid_loader()
        if use_tqdm:
            loader = tqdm.tqdm(loader, desc="validate")

        metric = self.validate_(loader)

        origin = col_name
        suffix = "validate"
        if kwargs["apply"] == "local":
            suffix += "_local"
        else:
            suffix += "_agg"
        tags = ("metric",)
        tags = change_tags(tags, add_field=suffix)
        # TODO figure out a better way to pass in metric for this pytorch
        #  validate function
        output_tensor_dict = {TensorKey(metric.name, origin, round_num, True, tags): metric.value}

        # Empty list represents metrics that should only be stored locally
        return output_tensor_dict, {}

    def train_task(
        self,
        col_name,
        round_num,
        input_tensor_dict,
        use_tqdm=False,
        epochs=1,
        **kwargs,
    ):
        """Train batches task.

        Train the model on the requested number of batches.

        Args:
            col_name (str): Name of the collaborator.
            round_num (int): What round is it.
            input_tensor_dict (dict): Required input tensors (for model).
            use_tqdm (bool): Use tqdm to print a progress bar (Default=True).
            epochs (int): The number of epochs to train.
            **kwargs: Additional parameters.

        Returns:
            global_output_dict (dict):  Tensors to send back to the aggregator.
            local_output_dict (dict):   Tensors to maintain in the local
                TensorDB.
        """
        self.rebuild_model(round_num, input_tensor_dict)
        # set to "training" mode
        self.train()
        self.to(self.device)
        for epoch in range(epochs):
            logger.info(f"Run {epoch} epoch of {round_num} round")
            loader = self.data_loader.get_train_loader()
            if use_tqdm:
                loader = tqdm.tqdm(loader, desc="train epoch")
            metric = self.train_(loader)
        # Output metric tensors (scalar)
        origin = col_name
        tags = ("trained",)
        output_metric_dict = {
            TensorKey(metric.name, origin, round_num, True, ("metric",)): metric.value
        }

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict(
            with_opt_vars=(self.opt_treatment == "CONTINUE_GLOBAL")
        )
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )

        # Create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in global_model_dict.items()
        }
        # Create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in local_model_dict.items()
        }
        # The train/validate aggregated function of the next round will look
        # for the updated model parameters.
        # This ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num + 1, False, ("model",)): nparray
            for tensor_name, nparray in local_model_dict.items()
        }

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict,
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict,
        }

        # Update the required tensors if they need to be pulled from the
        # aggregator
        # TODO this logic can break if different collaborators have different
        # roles between rounds.
        # For example, if a collaborator only performs validation in the first
        # round but training in the second, it has no way of knowing the
        # optimizer state tensor names to request from the aggregator because
        # these are only created after training occurs. A work around could
        # involve doing a single epoch of training on random data to get the
        # optimizer names, and then throwing away the model.
        if self.opt_treatment == "CONTINUE_GLOBAL":
            self.initialize_tensorkeys_for_functions(with_opt_vars=True)

        # This will signal that the optimizer values are now present,
        # and can be loaded when the model is rebuilt
        self.training_round_completed = True

        # Return global_tensor_dict, local_tensor_dict
        return global_tensor_dict, local_tensor_dict

    def get_tensor_dict(self, with_opt_vars=False):
        """Return the tensor dictionary.

        Args:
            with_opt_vars (bool): Return the tensor dictionary including the
                optimizer tensors (Default=False)

        Returns:
            state (dict): Tensor dictionary {**dict, **optimizer_dict}
        """
        # Gets information regarding tensor model layers and optimizer state.
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        # simple assignment is better
        # for now, state dict gives us names which is good
        # FIXME: do both and sanity check each time?

        state = to_cpu_numpy(self.state_dict())

        if with_opt_vars:
            opt_state = _get_optimizer_state(self.optimizer)
            state = {**state, **opt_state}

        return state

    def _get_weights_names(self, with_opt_vars=False):
        """Get information regarding tensor model layers and optimizer state.

        Args:
            with_opt_vars (bool): Flag to check if optimizer variables are
                included (Default=False).

        Returns:
            state (list): List of state keys.
        """
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        # simple assignment is better
        # for now, state dict gives us names which is good
        # FIXME: do both and sanity check each time?

        state = self.state_dict().keys()

        if with_opt_vars:
            opt_state = _get_optimizer_state(self.optimizer)
            state += opt_state.keys()

        return state

    def set_tensor_dict(self, tensor_dict, with_opt_vars=False):
        """Set the tensor dictionary.

        Args:
            tensor_dict (dict): The tensor dictionary.
            with_opt_vars (bool): Return the tensor dictionary including the
                optimizer tensors (Default=False).
        """
        # Sets tensors for model layers and optimizer state.
        # FIXME: self.parameters() instead? Unclear if load_state_dict() or
        #  simple assignment is better
        # for now, state dict gives us names, which is good
        # FIXME: do both and sanity check each time?

        # get device for correct placement of tensors
        device = self.device

        new_state = {}
        # Grabbing keys from model's state_dict helps to confirm we have
        # everything
        for k in self.state_dict():
            new_state[k] = torch.tensor(tensor_dict.pop(k)).to(device)

        # set model state
        self.load_state_dict(new_state)

        if with_opt_vars:
            # see if there is state to restore first
            if tensor_dict.pop("__opt_state_needed") == "true":
                _set_optimizer_state(self.get_optimizer(), device, tensor_dict)

            # sanity check that we did not record any state that was not used
            assert len(tensor_dict) == 0

    def get_optimizer(self):
        """Get the optimizer of this instance.

        Returns:
            Optimizer: The optimizer of the instance.
        """
        return self.optimizer

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """Get the required tensors for specified function that could be called
        as part of a task. By default, this is just all of the layers and
        optimizer of the model.

        Args:
            func_name (str): The function name.

        Returns:
            list : [TensorKey].
        """
        if func_name == "validate_task":
            local_model = "apply=" + str(kwargs["apply"])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Set the required tensors for all publicly accessible task methods.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function.

        Args:
            with_opt_vars (bool): Flag to check if optimizer variables are
                included. Defaults to False.

        Returns:
            None
        """
        # TODO there should be a way to programmatically iterate through
        #  all of the methods in the class and declare the tensors.
        # For now this is done manually

        output_model_dict = self.get_tensor_dict(with_opt_vars=with_opt_vars)
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )
        if not with_opt_vars:
            global_model_dict_val = global_model_dict
            local_model_dict_val = local_model_dict
        else:
            output_model_dict = self.get_tensor_dict(with_opt_vars=False)
            global_model_dict_val, local_model_dict_val = split_tensor_dict_for_holdouts(
                output_model_dict,
                **self.tensor_dict_split_fn_kwargs,
            )

        self.required_tensorkeys_for_function["train_task"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train_task"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        self.required_tensorkeys_for_function["train_task"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train_task"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function["validate_task"] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function["validate_task"]["apply=local"] = [
            TensorKey(tensor_name, "LOCAL", 0, False, ("trained",))
            for tensor_name in {**global_model_dict_val, **local_model_dict_val}
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict_val
        ]

    def load_native(
        self,
        filepath,
        model_state_dict_key="model_state_dict",
        optimizer_state_dict_key="optimizer_state_dict",
        **kwargs,
    ):
        """Load model and optimizer states from a pickled file specified by
        filepath. model_/optimizer_state_dict args can be specified if needed.
        Uses pt.load().

        Args:
            filepath (str): Path to pickle file created by pt.save().
            model_state_dict_key (str): key for model state dict in pickled
                file.
            optimizer_state_dict_key (str): key for optimizer state dict in
                picked file.
            **kwargs: Additional parameters.

        Returns:
            None
        """
        pickle_dict = torch.load(filepath, weights_only=True)  # nosec B614

        if model_state_dict_key in pickle_dict and optimizer_state_dict_key in pickle_dict:
            self.load_state_dict(pickle_dict[model_state_dict_key])
            self.optimizer.load_state_dict(pickle_dict[optimizer_state_dict_key])
        else:
            self.load_state_dict(pickle_dict)

    def save_native(
        self,
        filepath,
        model_state_dict_key="model_state_dict",
        optimizer_state_dict_key="optimizer_state_dict",
        **kwargs,
    ):
        """Save model and optimizer states in a picked file specified by the
        filepath. model_/optimizer_state_dicts are stored in the keys provided.
        Uses torch.save().

        Args:
            filepath (str): Path to pickle file to be created by torch.save().
                By default, model will be saved as `*.pt`
            model_state_dict_key (str): key for model state dict in pickled
                file.
            optimizer_state_dict_key (str): key for optimizer state dict in
                picked file.
            **kwargs: Additional parameters.

        Returns:
            None
        """
        pickle_dict = {
            model_state_dict_key: self.state_dict(),
            optimizer_state_dict_key: self.optimizer.state_dict(),
        }
        if "." not in str(filepath).split(os.sep)[-1]:
            filepath = str(filepath) + ".pt"

        torch.save(pickle_dict, filepath)
        return filepath

    def reset_opt_vars(self):
        """Reset optimizer variables.

        Resets the optimizer state variables.

        Returns:
            None
        """
        pass

    def train_(self, train_dataloader: Iterator[Tuple[np.ndarray, np.ndarray]]) -> Metric:
        """Train single epoch.

        Override this function in order to use custom training.

        Args:
            batch_generator (Iterator): Train dataset batch generator. Yields
                (samples, targets) tuples of
                size = `self.data_loader.batch_size`.

        Returns:
            Metric: An object containing name and np.ndarray value.
        """
        losses = []
        for data, target in train_dataloader:
            data, target = torch.tensor(data).to(self.device), torch.tensor(target).to(self.device)
            self.optimizer.zero_grad()
            output = self(data)
            loss = self.loss_fn(output=output, target=target)
            loss.backward()
            self.optimizer.step()
            losses.append(loss.detach().cpu().numpy())
        loss = np.mean(losses)
        return Metric(name=self.loss_fn.__name__, value=np.array(loss))

    def validate_(self, validation_dataloader: Iterator[Tuple[np.ndarray, np.ndarray]]) -> Metric:
        """
        Perform validation on PyTorch Model

        Override this function for your own custom validation function

        Args:
            validation_data_loader: Validation dataset batch generator.
                                    Yields (samples, targets) tuples.
        Returns:
            Metric: An object containing name and np.ndarray value
        """

        total_samples = 0
        val_score = 0
        with torch.no_grad():
            for data, target in validation_dataloader:
                samples = target.shape[0]
                total_samples += samples
                data, target = (
                    torch.tensor(data).to(self.device),
                    torch.tensor(target).to(self.device, dtype=torch.int64),
                )
                output = self(data)
                # get the index of the max log-probability
                pred = output.argmax(dim=1)
                val_score += pred.eq(target).sum().cpu().numpy()

        accuracy = val_score / total_samples
        return Metric(name="accuracy", value=np.array(accuracy))


def _derive_opt_state_dict(opt_state_dict):
    """Separate optimizer tensors from the tensor dictionary.

    Flattens the optimizer state dict so as to have key, value pairs with
    values as numpy arrays.
    The keys have sufficient info to restore opt_state_dict using
    expand_derived_opt_state_dict.

    Args:
        opt_state_dict (dict): The optimizer state dictionary.

    Returns:
        derived_opt_state_dict (dict): Derived optimizer state dictionary.
    """
    derived_opt_state_dict = {}

    # Determine if state is needed for this optimizer.
    if len(opt_state_dict["state"]) == 0:
        derived_opt_state_dict["__opt_state_needed"] = "false"
        return derived_opt_state_dict

    derived_opt_state_dict["__opt_state_needed"] = "true"

    # Using one example state key, we collect keys for the corresponding
    # dictionary value.
    example_state_key = opt_state_dict["param_groups"][0]["params"][0]
    example_state_subkeys = set(opt_state_dict["state"][example_state_key].keys())

    # We assume that the state collected for all params in all param groups is
    # the same.
    # We also assume that whether or not the associated values to these state
    # subkeys is a tensor depends only on the subkey.
    # Using assert statements to break the routine if these assumptions are
    # incorrect.
    for state_key in opt_state_dict["state"].keys():
        assert example_state_subkeys == set(opt_state_dict["state"][state_key].keys())
        for state_subkey in example_state_subkeys:
            assert isinstance(
                opt_state_dict["state"][example_state_key][state_subkey],
                torch.Tensor,
            ) == isinstance(opt_state_dict["state"][state_key][state_subkey], torch.Tensor)

    state_subkeys = list(opt_state_dict["state"][example_state_key].keys())

    # Tags will record whether the value associated to the subkey is a
    # tensor or not.
    state_subkey_tags = []
    for state_subkey in state_subkeys:
        if isinstance(
            opt_state_dict["state"][example_state_key][state_subkey],
            torch.Tensor,
        ):
            state_subkey_tags.append("istensor")
        else:
            state_subkey_tags.append("")
    state_subkeys_and_tags = list(zip(state_subkeys, state_subkey_tags))

    # Forming the flattened dict, using a concatenation of group index,
    # subindex, tag, and subkey inserted into the flattened dict key -
    # needed for reconstruction.
    nb_params_per_group = []
    for group_idx, group in enumerate(opt_state_dict["param_groups"]):
        for idx, param_id in enumerate(group["params"]):
            for subkey, tag in state_subkeys_and_tags:
                if tag == "istensor":
                    new_v = opt_state_dict["state"][param_id][subkey].cpu().numpy()
                else:
                    new_v = np.array([opt_state_dict["state"][param_id][subkey]])
                derived_opt_state_dict[f"__opt_state_{group_idx}_{idx}_{tag}_{subkey}"] = new_v
        nb_params_per_group.append(idx + 1)
    # group lengths are also helpful for reconstructing
    # original opt_state_dict structure
    derived_opt_state_dict["__opt_group_lengths"] = np.array(nb_params_per_group)

    return derived_opt_state_dict


def expand_derived_opt_state_dict(derived_opt_state_dict, device):
    """Expand the optimizer state dictionary.

    Takes a derived opt_state_dict and creates an opt_state_dict suitable as
    input for load_state_dict for restoring optimizer state.

    Reconstructing state_subkeys_and_tags using the example key
    prefix, "__opt_state_0_0_", certain to be present.

    Args:
        derived_opt_state_dict (dict): Optimizer state dictionary.
        device (str): The device to be used.

    Returns:
        opt_state_dict (dict): Optimizer state dictionary.
    """
    state_subkeys_and_tags = []
    for key in derived_opt_state_dict:
        if key.startswith("__opt_state_0_0_"):
            stripped_key = key[16:]
            if stripped_key.startswith("istensor_"):
                this_tag = "istensor"
                subkey = stripped_key[9:]
            else:
                this_tag = ""
                subkey = stripped_key[1:]
            state_subkeys_and_tags.append((subkey, this_tag))

    opt_state_dict = {"param_groups": [], "state": {}}
    nb_params_per_group = list(derived_opt_state_dict.pop("__opt_group_lengths").astype(np.int32))

    # Construct the expanded dict.
    for group_idx, nb_params in enumerate(nb_params_per_group):
        these_group_ids = [f"{group_idx}_{idx}" for idx in range(nb_params)]
        opt_state_dict["param_groups"].append({"params": these_group_ids})
        for this_id in these_group_ids:
            opt_state_dict["state"][this_id] = {}
            for subkey, tag in state_subkeys_and_tags:
                flat_key = f"__opt_state_{this_id}_{tag}_{subkey}"
                if tag == "istensor":
                    new_v = torch.tensor(derived_opt_state_dict.pop(flat_key))
                else:
                    # Here (for currently supported optimizers) the subkey
                    # should be 'step' and the length of array should be one.
                    assert subkey == "step"
                    assert len(derived_opt_state_dict[flat_key]) == 1
                    new_v = int(derived_opt_state_dict.pop(flat_key))
                opt_state_dict["state"][this_id][subkey] = new_v

    # sanity check that we did not miss any optimizer state
    assert len(derived_opt_state_dict) == 0

    return opt_state_dict


def _get_optimizer_state(optimizer):
    """Return the optimizer state.

    Args:
        optimizer (Optimizer): The optimizer for the model.

    Returns:
        derived_opt_state_dict (dict): Derived optimizer state dictionary.
    """
    opt_state_dict = deepcopy(optimizer.state_dict())

    # Optimizer state might not have some parts representing frozen parameters
    # So we do not synchronize them
    param_keys_with_state = set(opt_state_dict["state"].keys())
    for group in opt_state_dict["param_groups"]:
        local_param_set = set(group["params"])
        params_to_sync = local_param_set & param_keys_with_state
        group["params"] = sorted(params_to_sync)

    derived_opt_state_dict = _derive_opt_state_dict(opt_state_dict)

    return derived_opt_state_dict


def _set_optimizer_state(optimizer, device, derived_opt_state_dict):
    """Set the optimizer state.

    Args:
        optimizer (Optimizer): The optimizer for the model.
        device (str): The device to be used.
        derived_opt_state_dict (dict): Derived optimizer state dictionary.

    Returns:
        None
    """
    temp_state_dict = expand_derived_opt_state_dict(derived_opt_state_dict, device)

    # FIXME: Figure out whether or not this breaks learning rate
    #  scheduling and the like.
    # Setting default values.
    # All optimizer.defaults are considered as not changing over course of
    # training.
    for group in temp_state_dict["param_groups"]:
        for k, v in optimizer.defaults.items():
            group[k] = v

    optimizer.load_state_dict(temp_state_dict)


def to_cpu_numpy(state):
    """Send data to CPU as Numpy array.

    Args:
        state (dict): The state dictionary.

    Returns:
        state (dict): State dictionary with values as numpy arrays.
    """
    # deep copy so as to decouple from active model
    state = deepcopy(state)

    for k, v in state.items():
        # When restoring, we currently assume all values are tensors.
        if not torch.is_tensor(v):
            raise ValueError(
                "We do not currently support non-tensors coming from model.state_dict()"
            )
        # get as a numpy array, making sure is on cpu
        state[k] = v.cpu().numpy()
    return state



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Pipeline module."""

import numpy as np


class Transformer:
    """Base class for data transformation.

    This class defines the basic structure of a data transformer. It should be
    subclassed when implementing new types of data transformations.
    """

    def forward(self, data, **kwargs):
        """Forward pass data transformation.

        Implement the data transformation.
        This method should be overridden by all subclasses.

        Args:
            data: The data to be transformed.
            **kwargs: Additional parameters to pass to the function

        Returns:
            transformed_data: The transformed data.
            metadata: The metadata for the transformation.
        """
        raise NotImplementedError

    def backward(self, data, metadata, **kwargs):
        """Backward pass data transformation.

        Implement the data transformation needed when going the opposite
        direction to the forward method.
        This method should be overridden by all subclasses.

        Args:
            data: The transformed data.
            metadata: The metadata for the transformation.
            **kwargs: Additional keyword arguments for the transformation.

        Returns:
            transformed_data: The original data before the transformation.
        """
        raise NotImplementedError


class Float32NumpyArrayToBytes(Transformer):
    """Transformer class for converting float32 Numpy arrays to bytes
    arrays."""

    def __init__(self):
        """Initialize Float32NumpyArrayToBytes."""
        self.lossy = False

    def forward(self, data: np.ndarray, **kwargs):
        """Convert a float32 Numpy array to bytes.

        Args:
            data: The float32 Numpy array to be converted.
            **kwargs: Additional keyword arguments for the conversion.

        Returns:
            data_bytes: The data converted to bytes.
            metadata: The metadata for the conversion.
        """
        # TODO: Warn when this casting is being performed.
        if data.dtype != np.float32:
            data = data.astype(np.float32)
        array_shape = data.shape
        # Better call it array_shape?
        metadata = {"int_list": list(array_shape)}
        data_bytes = data.tobytes(order="C")
        return data_bytes, metadata

    def backward(self, data, metadata, **kwargs):
        """Convert bytes back to a float32 Numpy array.

        Args:
            data: The data in bytes.
            metadata: The metadata for the conversion.

        Returns:
            The data converted back to a float32 Numpy array.
        """
        array_shape = tuple(metadata["int_list"])
        flat_array = np.frombuffer(data, dtype=np.float32)
        # For integer parameters we probably should unpack arrays
        # with shape (1,)
        return np.reshape(flat_array, newshape=array_shape, order="C")


class TransformationPipeline:
    """Data Transformer Pipeline Class.

    This class is a pipeline of transformers that transform data in a
    sequential manner.

    A sequential pipeline to transform (e.x. compress) data (e.x. layer of
    model_weights) as well as return metadata (if needed) for the
    reconstruction process carried out by the backward method.

    Attributes:
        transformers (list): The list of transformers in the pipeline.
    """

    def __init__(self, transformers, **kwargs):
        """Initialize TransformationPipeline.

        Args:
            transformers (list): The list of transformers in the pipeline.
            **kwargs: Additional keyword arguments for the pipeline.
        """
        self.transformers = transformers

    def forward(self, data, **kwargs):
        """Forward pass of pipeline data transformer.

        Args:
            data: The data to be transformed.
            **kwargs: Additional keyword arguments for the transformation.

        Returns:
            data: The transformed data.
            transformer_metadata: The metadata for the transformation.
        """
        # dataformat::numpy::float.32
        # model proto:: a collection of tensor_dict proto
        # protobuff::-> a layer of weights
        # input::tensor_dict:{"layer1":np.array(float32, [128,3,28,28]),
        # "layer2": np.array()}
        # output::meta data::numpy::int array
        # (data, transformer_metadata)::(float32, dictionary o
        # key+float32 values)
        # input:: numpy_data (float32)
        # input:: (data(bytes), transformer_metadata_list::a list of dictionary
        # from int to float)

        transformer_metadata = []
        for transformer in self.transformers:
            data, metadata = transformer.forward(data=data, **kwargs)
            transformer_metadata.append(metadata)
        return data, transformer_metadata

    def backward(self, data, transformer_metadata, **kwargs):
        """Backward pass of pipeline data transformer.

        Args:
            data: The transformed data.
            transformer_metadata: The metadata for the transformation.
            **kwargs: Additional keyword arguments for the transformation.

        Returns:
            The original data before the transformation.
        """
        for transformer in self.transformers[::-1]:
            data = transformer.backward(data=data, metadata=transformer_metadata.pop(), **kwargs)
        return data

    def is_lossy(self):
        """If any of the transformers are lossy, then the pipeline is lossy.

        Returns:
            True if any of the transformers in the pipeline are lossy, False
                otherwise.
        """
        return any(transformer.lossy for transformer in self.transformers)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Experimental Director module."""

import asyncio
import logging
import time
from collections import defaultdict
from pathlib import Path
from typing import Any, AsyncGenerator, Dict, Iterable, Optional, Tuple, Union

import dill

from openfl.experimental.workflow.component.director.experiment import (
    Experiment,
    ExperimentsRegistry,
)
from openfl.experimental.workflow.transport.grpc.exceptions import EnvoyNotFoundError

logger = logging.getLogger(__name__)


class Director:
    """Director class for managing experiments and envoys.

    Attributes:
        tls (bool): A flag indicating if TLS should be used for connections.
        root_certificate (Optional[Union[Path, str]]): The path to the root certificate
            for TLS.
        private_key (Optional[Union[Path, str]]): The path to the private key for TLS.
        certificate (Optional[Union[Path, str]]): The path to the certificate for TLS.
        director_config (Optional[Path]): Path to director_config file
        install_requirements (bool): A flag indicating if the requirements
            should be installed.
        _flow_status (Queue): Stores the flow status
        experiments_registry (ExperimentsRegistry): An object of
            ExperimentsRegistry to store the experiments.
        col_exp (dict): A dictionary to store the experiments for
            collaborators.
        col_exp_queues (defaultdict): A defaultdict to store the experiment
            queues for collaborators.
        _envoy_registry (dict): A dcitionary to store envoy info
        envoy_health_check_period (int): The period for health check of envoys
            in seconds.
        authorized_cols (list): A list of authorized envoys
    """

    def __init__(
        self,
        *,
        tls: bool = True,
        root_certificate: Optional[Union[Path, str]] = None,
        private_key: Optional[Union[Path, str]] = None,
        certificate: Optional[Union[Path, str]] = None,
        director_config: Optional[Path] = None,
        envoy_health_check_period: int = 60,
        install_requirements: bool = True,
    ) -> None:
        """Initialize a Director object.

        Args:
            tls (bool, optional): A flag indicating if TLS should be used for
                connections. Defaults to True.
            root_certificate (Optional[Union[Path, str]]): The path to the
                root certificate for TLS. Defaults to None.
            private_key (Optional[Union[Path, str]]): The path to the private
                key for TLS. Defaults to None.
            certificate (Optional[Union[Path, str]]): The path to the
                certificate for TLS. Defaults to None.
            director_config (Optional[Path]): Path to director_config file
            envoy_health_check_period (int): The period for health check of envoys
            in seconds.
            install_requirements (bool, optional): A flag indicating if the
                requirements should be installed. Defaults to True.
        """
        self.tls = tls
        self.root_certificate = root_certificate
        self.private_key = private_key
        self.certificate = certificate
        self.director_config = director_config
        self.install_requirements = install_requirements
        self._flow_status = asyncio.Queue()

        self.experiments_registry = ExperimentsRegistry()
        self.col_exp = {}
        self.col_exp_queues = defaultdict(asyncio.Queue)
        self._envoy_registry = {}
        self.envoy_health_check_period = envoy_health_check_period
        # authorized_cols refers to envoy & collaborator pair (one to one mapping)
        self.authorized_cols = []

    async def start_experiment_execution_loop(self) -> None:
        """Run tasks and experiments here"""
        loop = asyncio.get_event_loop()
        while True:
            try:
                async with self.experiments_registry.get_next_experiment() as experiment:
                    await self._wait_for_authorized_envoys()
                    run_aggregator_future = loop.create_task(
                        experiment.start(
                            root_certificate=self.root_certificate,
                            certificate=self.certificate,
                            private_key=self.private_key,
                            tls=self.tls,
                            director_config=self.director_config,
                            install_requirements=self.install_requirements,
                        )
                    )
                    # Adding the experiment to collaborators queues
                    for col_name in experiment.collaborators:
                        queue = self.col_exp_queues[col_name]
                        await queue.put(experiment.name)
                    # Wait for the experiment to complete and save the result
                    flow_status = await run_aggregator_future
                    await self._flow_status.put(flow_status)
            except Exception as e:
                logger.error(f"Error while executing experiment: {e}")
                raise

    async def _wait_for_authorized_envoys(self) -> None:
        """Wait until all authorized envoys are connected"""
        while not all(envoy in self.get_envoys().keys() for envoy in self.authorized_cols):
            connected_envoys = len(
                [envoy for envoy in self.authorized_cols if envoy in self.get_envoys().keys()]
            )
            logger.info(
                f"Waiting for {connected_envoys}/{len(self.authorized_cols)} "
                "authorized envoys to connect..."
            )
            await asyncio.sleep(10)

    async def get_flow_state(self) -> Tuple[bool, bytes]:
        """Wait until the experiment flow status indicates completion
        and return the status along with a serialized FLSpec object.

        Returns:
            status (bool): The flow status.
            flspec_obj (bytes): A serialized FLSpec object (in bytes) using dill.
        """
        status, flspec_obj = await self._flow_status.get()
        return status, dill.dumps(flspec_obj)

    async def wait_experiment(self, envoy_name: str) -> str:
        """Waits for an experiment to be ready for a given envoy.

        Args:
            envoy_name (str): The name of the envoy.

        Returns:
            str: The name of the experiment on the queue.
        """
        experiment_name = self.col_exp.get(envoy_name)
        # If any envoy gets disconnected
        if experiment_name and experiment_name in self.experiments_registry:
            experiment = self.experiments_registry[experiment_name]
            if experiment.aggregator.current_round < experiment.aggregator.rounds_to_train:
                return experiment_name

        self.col_exp[envoy_name] = None
        queue = self.col_exp_queues[envoy_name]
        experiment_name = await queue.get()
        self.col_exp[envoy_name] = experiment_name

        return experiment_name

    async def set_new_experiment(
        self,
        experiment_name: str,
        sender_name: str,
        collaborator_names: Iterable[str],
        experiment_archive_path: Path,
    ) -> bool:
        """Set new experiment.

        Args:
            experiment_name (str): String id for experiment.
            sender_name (str): The name of the sender.
            collaborator_names (Iterable[str]): Names of collaborators.
            experiment_archive_path (Path): Path of the experiment.

        Returns:
            bool : Boolean returned if the experiment register was successful.
        """
        experiment = Experiment(
            name=experiment_name,
            archive_path=experiment_archive_path,
            collaborators=collaborator_names,
            users=[sender_name],
            sender=sender_name,
        )

        self.authorized_cols = collaborator_names
        self.experiments_registry.add(experiment)
        return True

    async def stream_experiment_stdout(
        self, experiment_name: str, caller: str
    ) -> AsyncGenerator[Optional[Dict[str, Any]], None]:
        """Stream stdout from the aggregator.

        This method takes next stdout dictionary from the aggregator's queue
        and returns it to the caller.

        Args:
            experiment_name (str): String id for experiment.
            caller (str): String id for experiment owner.

        Yields:
            Optional[Dict[str, str]]: A dictionary containing the keys
            'stdout_origin', 'task_name', and 'stdout_value' if the queue is not empty,
            or None if the queue is empty but the experiment is still running.
        """
        if (
            experiment_name not in self.experiments_registry
            or caller not in self.experiments_registry[experiment_name].users
        ):
            raise Exception(
                f'No experiment name "{experiment_name}" in experiments list, or caller "{caller}"'
                f" does not have access to this experiment"
            )
        while not self.experiments_registry[experiment_name].aggregator:
            await asyncio.sleep(5)
        aggregator = self.experiments_registry[experiment_name].aggregator
        while True:
            if not aggregator.stdout_queue.empty():
                # Yield the next item from the queue
                yield aggregator.stdout_queue.get()
            elif aggregator.all_quit_jobs_sent():
                # Stop Iteration if all jobs have quit and the queue is empty
                break
            else:
                # Yield none if the queue is empty but the experiment is still running.
                yield None

    def get_experiment_data(self, experiment_name: str) -> Path:
        """Get experiment data.

        Args:
            experiment_name (str): String id for experiment.

        Returns:
            str: Path of archive.
        """
        return self.experiments_registry[experiment_name].archive_path

    def ack_envoy_connection_request(self, envoy_name: str) -> bool:
        """Save the envoy info into _envoy_registry

        Args:
            envoy_name (str): Name of the envoy

        Returns:
            bool: Always returns True to indicate the envoy
                has been successfully acknowledged.
        """
        self._envoy_registry[envoy_name] = {
            "name": envoy_name,
            "is_online": True,
            "is_experiment_running": False,
            "last_updated": time.time(),
            "valid_duration": 2 * self.envoy_health_check_period,
        }
        # Currently always returns True, indicating the envoy was added successfully.
        # Future logic might change this to handle conditions.
        return True

    def get_envoys(self) -> Dict[str, Any]:
        """Gets list of connected envoys

        Returns:
            dict: Dictionary with the status information about envoys.
        """
        logger.debug("Envoy registry: %s", self._envoy_registry)
        for envoy in self._envoy_registry.values():
            envoy["is_online"] = time.time() < envoy.get("last_updated", 0) + envoy.get(
                "valid_duration", 0
            )
            envoy["experiment_name"] = self.col_exp.get(envoy["name"], "None")

        return self._envoy_registry

    def update_envoy_status(
        self,
        *,
        envoy_name: str,
        is_experiment_running: bool,
    ) -> int:
        """Accept health check from envoy.

        Args:
            envoy_name (str): String id for envoy.
            is_experiment_running (bool): Boolean value for the status of the
                experiment.

        Raises:
            EnvoyNotFoundError: When Unknown envoy {envoy_name}.

        Returns:
            int: Value of the envoy_health_check_period.
        """
        envoy_info = self._envoy_registry.get(envoy_name)
        if not envoy_info:
            logger.error(f"Unknown envoy {envoy_name}")
            raise EnvoyNotFoundError(f"Unknown envoy {envoy_name}")

        envoy_info.update(
            {
                "is_online": True,
                "is_experiment_running": is_experiment_running,
                "valid_duration": 2 * self.envoy_health_check_period,
                "last_updated": time.time(),
            }
        )

        return self.envoy_health_check_period



# Copyright 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""AggregatorGRPCServer module."""

import logging
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from random import random
from time import sleep

from grpc import StatusCode, server, ssl_server_credentials

from openfl.experimental.workflow.protocols import aggregator_pb2, aggregator_pb2_grpc
from openfl.experimental.workflow.transport.grpc.grpc_channel_options import channel_options
from openfl.protocols.utils import datastream_to_proto, proto_to_datastream

logger = logging.getLogger(__name__)


class AggregatorGRPCServer(aggregator_pb2_grpc.AggregatorServicer):
    """GRPC server class for the Aggregator."""

    def __init__(
        self,
        aggregator,
        agg_port,
        tls=True,
        disable_client_auth=False,
        root_certificate=None,
        certificate=None,
        private_key=None,
        **kwargs,
    ):
        """Class initializer.

        Args:
            aggregator: The aggregator
        Args:
            fltask (FLtask): The gRPC service task.
            tls (bool): To disable the TLS. (Default: True)
            disable_client_auth (bool): To disable the client side
            authentication. (Default: False)
            root_certificate (str): File path to the CA certificate.
            certificate (str): File path to the server certificate.
            private_key (str): File path to the private key.
            kwargs (dict): Additional arguments to pass into function
        """
        self.aggregator = aggregator
        self.uri = f"[::]:{agg_port}"
        self.tls = tls
        self.disable_client_auth = disable_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key
        self.server = None
        self.server_credentials = None

        self.logger = logging.getLogger(__name__)

    def validate_collaborator(self, request, context):
        """Validate the collaborator.

        Args:
            request: The gRPC message request
            context: The gRPC context

        Raises:
            ValueError: If the collaborator or collaborator certificate is not
                valid then raises error.
        """
        if self.tls:
            common_name = context.auth_context()["x509_common_name"][0].decode("utf-8")
            collaborator_common_name = request.header.sender
            if not self.aggregator.valid_collaborator_cn_and_id(
                common_name, collaborator_common_name
            ):
                # Random delay in authentication failures
                sleep(5 * random())
                context.abort(
                    StatusCode.UNAUTHENTICATED,
                    f"Invalid collaborator. CN: |{common_name}| "
                    f"collaborator_common_name: |{collaborator_common_name}|",
                )

    def get_header(self, collaborator_name):
        """Compose and return MessageHeader.

        Args:
            collaborator_name : str
                The collaborator the message is intended for
        """
        return aggregator_pb2.MessageHeader(
            sender=self.aggregator.uuid,
            receiver=collaborator_name,
            federation_uuid=self.aggregator.federation_uuid,
            single_col_cert_common_name=self.aggregator.single_col_cert_common_name,
        )

    def check_request(self, request):
        """Validate request header matches expected values.

        Args:
            request : protobuf
                Request sent from a collaborator that requires validation
        """
        assert request.header.sender in self.aggregator.authorized_cols, (
            f"Sender in request header is not authorized. "
            f"Expected: one of {self.aggregator.authorized_cols}, Actual: {request.header.sender}"
        )

        assert request.header.receiver == self.aggregator.uuid, (
            f"Receiver in request header does not match aggregator UUID. "
            f"Expected: {self.aggregator.uuid}, Actual: {request.header.receiver}"
        )

        assert request.header.federation_uuid == self.aggregator.federation_uuid, (
            f"Federation UUID in request header does not match. "
            f"Expected: {self.aggregator.federation_uuid}, Actual: {request.header.federation_uuid}"
        )

        assert (
            request.header.single_col_cert_common_name
            == self.aggregator.single_col_cert_common_name
        ), (
            f"Single collaborator certificate common name in request header does not match. "
            f"Expected: {self.aggregator.single_col_cert_common_name}, Actual: {request.header.single_col_cert_common_name}"  # noqa: E501
        )

    def SendTaskResults(self, request, context):  # NOQA:N802
        """Processes a request from a collaborator to retrieve the results of a locally
        executed task.

        Args:
            request: The gRPC message request
            context: The gRPC context
        """
        proto = datastream_to_proto(aggregator_pb2.TaskResultsRequest(), request)
        self.validate_collaborator(proto, context)
        self.check_request(proto)
        collaborator_name = proto.header.sender
        round_number = (proto.round_number,)
        next_step = (proto.next_step,)
        execution_environment = proto.execution_environment

        _ = self.aggregator.send_task_results(
            collaborator_name, round_number[0], next_step, execution_environment
        )

        return aggregator_pb2.TaskResultsResponse(header=self.get_header(collaborator_name))

    def GetTasks(self, request, context):  # NOQA:N802
        """Request a job from aggregator.

        Args:
            request: The gRPC message request
            context: The gRPC context
        """
        self.validate_collaborator(request, context)
        self.check_request(request)
        collaborator_name = request.header.sender

        rn, f, ee, st, q = self.aggregator.get_tasks(request.header.sender)

        response = aggregator_pb2.GetTasksResponse(
            header=self.get_header(collaborator_name),
            round_number=rn,
            function_name=f,
            execution_environment=ee,
            sleep_time=st,
            quit=q,
        )
        return proto_to_datastream(response)

    def CallCheckpoint(self, request, context):  # NOQA:N802
        """Request aggregator to perform a checkpoint for a given function.

        Args:
            request: The gRPC message request
            context: The gRPC context
        """
        proto = datastream_to_proto(aggregator_pb2.CheckpointRequest(), request)
        self.validate_collaborator(proto, context)
        self.check_request(proto)
        collaborator_name = proto.header.sender
        execution_environment = proto.execution_environment
        function = proto.function
        stream_buffer = proto.stream_buffer

        self.aggregator.call_checkpoint(
            collaborator_name, execution_environment, function, stream_buffer
        )

        return aggregator_pb2.CheckpointResponse(header=self.get_header(collaborator_name))

    def get_server(self):
        """Return gRPC server."""
        self.server = server(ThreadPoolExecutor(max_workers=cpu_count()), options=channel_options)

        aggregator_pb2_grpc.add_AggregatorServicer_to_server(self, self.server)

        if not self.tls:
            self.logger.warn("gRPC is running on insecure channel with TLS disabled.")
            port = self.server.add_insecure_port(self.uri)
            self.logger.info(f"Insecure port: {port}")

        else:
            with open(self.private_key, "rb") as f:
                private_key_b = f.read()
            with open(self.certificate, "rb") as f:
                certificate_b = f.read()
            with open(self.root_certificate, "rb") as f:
                root_certificate_b = f.read()

            if self.disable_client_auth:
                self.logger.warn("Client-side authentication is disabled.")

            self.server_credentials = ssl_server_credentials(
                ((private_key_b, certificate_b),),
                root_certificates=root_certificate_b,
                require_client_auth=not self.disable_client_auth,
            )

            self.server.add_secure_port(self.uri, self.server_credentials)

        return self.server



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""TensorCodec module."""

import numpy as np

from openfl.pipelines import NoCompressionPipeline
from openfl.utilities import TensorKey, change_tags


class TensorCodec:
    """TensorCodec is responsible for the following.

    1. Tracking the compression/decompression related dependencies of a given
    tensor.
    2. Acting as a TensorKey aware wrapper for the compression_pipeline
    functionality.

    Attributes:
        compression_pipeline: The pipeline used for compression.
        lossless_pipeline: The pipeline used for lossless compression.
    """

    def __init__(self, compression_pipeline):
        """Initialize the TensorCodec.

        Args:
            compression_pipeline: The pipeline used for compression.
        """
        self.compression_pipeline = compression_pipeline
        if self.compression_pipeline.is_lossy():
            self.lossless_pipeline = NoCompressionPipeline()
        else:
            self.lossless_pipeline = compression_pipeline

    def set_lossless_pipeline(self, lossless_pipeline):
        """
        Set lossless pipeline.

        Args:
            lossless_pipeline: The pipeline to be set as the lossless pipeline.
                It should be a pipeline that is not lossy.

        Raises:
            AssertionError: If the provided pipeline is not lossless.
        """
        assert lossless_pipeline.is_lossy() is False, "The provided pipeline is not lossless"
        self.lossless_pipeline = lossless_pipeline

    def compress(self, tensor_key, data, require_lossless=False, **kwargs):
        """Function-wrapper around the tensor_pipeline.forward function.

        It also keeps track of the tensorkeys associated with the compressed
        nparray.

        Args:
            tensor_key: TensorKey is provided to verify it should be
                compressed, and new TensorKeys returned will be derivatives of
                the existing tensor_name.
            data: (uncompressed) numpy array associated with the tensor_key.
            require_lossless: boolean. Does tensor require compression.

        Returns:
            compressed_tensor_key: Tensorkey corresponding to the decompressed
                tensor.
            compressed_nparray: The compressed tensor.
            metadata: metadata associated with compressed tensor.
        """
        if require_lossless:
            compressed_nparray, metadata = self.lossless_pipeline.forward(data, **kwargs)
        else:
            compressed_nparray, metadata = self.compression_pipeline.forward(data, **kwargs)
        # Define the compressed tensorkey that should be
        # returned ('trained.delta'->'trained.delta.lossy_compressed')
        tensor_name, origin, round_number, report, tags = tensor_key
        if not self.compression_pipeline.is_lossy() or require_lossless:
            new_tags = change_tags(tags, add_field="compressed")
        else:
            new_tags = change_tags(tags, add_field="lossy_compressed")
        compressed_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
        return compressed_tensor_key, compressed_nparray, metadata

    def decompress(
        self,
        tensor_key,
        data,
        transformer_metadata,
        require_lossless=False,
        **kwargs,
    ):
        """
        Function-wrapper around the tensor_pipeline.backward function.

        It also keeps track of the tensorkeys associated with the decompressed
        nparray.

        Args:
            tensor_key: TensorKey is provided to verify it should be
                decompressed, and new TensorKeys returned will be derivatives
                of the existing tensor_name.
            data: (compressed) numpy array associated with the tensor_key.
            transformer_metadata: metadata associated with the compressed
                tensor.
            require_lossless: boolean, does data require lossless
                decompression.

        Returns:
            decompressed_tensor_key: Tensorkey corresponding to the
                decompressed tensor.
            decompressed_nparray: The decompressed tensor.
        """
        tensor_name, origin, round_number, report, tags = tensor_key

        assert len(transformer_metadata) > 0, "metadata must be included for decompression"
        assert ("compressed" in tags) or ("lossy_compressed" in tags), (
            "Cannot decompress an uncompressed tensor"
        )
        if require_lossless:
            assert "compressed" in tags, "Cannot losslessly decompress lossy tensor"

        if require_lossless or "compressed" in tags:
            decompressed_nparray = self.lossless_pipeline.backward(
                data, transformer_metadata, **kwargs
            )
        else:
            decompressed_nparray = self.compression_pipeline.backward(
                data, transformer_metadata, **kwargs
            )
        # Define the decompressed tensorkey that should be returned
        if "lossy_compressed" in tags:
            new_tags = change_tags(
                tags,
                add_field="lossy_decompressed",
                remove_field="lossy_compressed",
            )
            decompressed_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
        elif "compressed" in tags:
            # 'compressed' == lossless compression; no need for
            # compression related tag after decompression
            new_tags = change_tags(tags, remove_field="compressed")
            decompressed_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
        else:
            raise NotImplementedError("Decompression is only supported on compressed data")

        return decompressed_tensor_key, decompressed_nparray

    @staticmethod
    def generate_delta(tensor_key, nparray, base_model_nparray):
        """Create delta from the updated layer and base layer.

        Args:
            tensor_key: This is the tensor_key associated with the nparray.
                Should have a tag of 'trained' or 'aggregated'
            nparray: The nparray that corresponds to the tensorkey.
            base_model_nparray: The base model tensor that will be subtracted
                from the new weights.

        Returns:
            delta_tensor_key: Tensorkey that corresponds to the delta weight
                array.
            delta: Difference between the provided tensors.
        """
        tensor_name, origin, round_number, report, tags = tensor_key
        if not np.isscalar(nparray):
            assert nparray.shape == base_model_nparray.shape, (
                f"Shape of updated layer ({nparray.shape}) is not equal to base "
                f"layer shape of ({base_model_nparray.shape})"
            )
        assert "model" not in tags, (
            "The tensorkey should be provided from the layer with new weights, not the base model"
        )
        new_tags = change_tags(tags, add_field="delta")
        delta_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
        return delta_tensor_key, nparray - base_model_nparray

    @staticmethod
    def apply_delta(tensor_key, delta, base_model_nparray, creates_model=False):
        """Add delta to the nparray.

        Args:
            tensor_key: This is the tensor_key associated with the delta.
                Should have a tag of 'trained' or 'aggregated'.
            delta: Weight delta between the new model and old model.
            base_model_nparray: The nparray that corresponds to the prior
                weights.
            creates_model: If flag is set, the tensorkey returned will
                correspond to the aggregator model.

        Returns:
            new_model_tensor_key: Latest model layer tensorkey.
            new_model_nparray: Latest layer weights.
        """
        tensor_name, origin, round_number, report, tags = tensor_key
        if not np.isscalar(base_model_nparray):
            assert delta.shape == base_model_nparray.shape, (
                f"Shape of delta ({delta.shape}) is not equal to shape of model"
                f" layer ({base_model_nparray.shape})"
            )
        # assert('model' in tensor_key[3]), 'The tensorkey should be provided
        # from the base model'
        # Aggregator UUID has the prefix 'aggregator'
        if "aggregator" in origin and not creates_model:
            new_tags = change_tags(tags, remove_field="delta")
            new_model_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
        else:
            new_model_tensor_key = TensorKey(tensor_name, origin, round_number, report, ("model",))

        return new_model_tensor_key, base_model_nparray + delta

    def find_dependencies(self, tensor_key, send_model_deltas):
        """Resolve the tensors required to do the specified operation.

        Args:
            tensor_key: A tuple containing the tensor name, origin, round
                number, report, and tags.
            send_model_deltas: A boolean flag indicating whether to send model
                deltas.

        Returns:
            tensor_key_dependencies: A list of tensor keys that are
                dependencies of the given tensor key.
        """
        tensor_key_dependencies = []

        tensor_name, origin, round_number, report, tags = tensor_key

        if "model" in tags and send_model_deltas:
            if round_number >= 1:
                # The new model can be generated by previous model + delta
                tensor_key_dependencies.append(
                    TensorKey(tensor_name, origin, round_number - 1, report, tags)
                )
                if self.compression_pipeline.is_lossy():
                    new_tags = ("aggregated", "delta", "lossy_compressed")
                else:
                    new_tags = ("aggregated", "delta", "compressed")
                tensor_key_dependencies.append(
                    TensorKey(tensor_name, origin, round_number, report, new_tags)
                )

        return tensor_key_dependencies



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PyTorchDataLoader module."""

from openfl.federated.data.loader import DataLoader


class GaNDLFDataLoaderWrapper(DataLoader):
    """A class used to represent a data loader for the Generally Nuanced Deep
    Learning Framework (GaNDLF).

    Attributes:
        train_csv (str): Path to the training CSV file.
        val_csv (str): Path to the validation CSV file.
        train_dataloader (DataLoader): DataLoader object for the training data.
        val_dataloader (DataLoader): DataLoader object for the validation data.
        feature_shape (tuple): Shape of an example feature array.
    """

    def __init__(self, data_path=None, **kwargs):
        """Initializes the GaNDLFDataLoaderWrapper object.

        Args:
            data_path (str, optional): The path to the directory containing the data.
                If None, initialize for model creation only.
            **kwargs: Additional arguments to pass to the function.
        """
        self.train_csv = None
        self.val_csv = None
        self.train_dataloader = None
        self.val_dataloader = None

        # If data_path is None, this is being used for model initialization only
        if data_path is None:
            return

        # Otherwise set up paths for actual data loading
        if "inference" in data_path:
            self.train_csv = None
        else:
            self.train_csv = data_path + "/train.csv"
        self.val_csv = data_path + "/valid.csv"

    def set_dataloaders(self, train_dataloader, val_dataloader):
        """Sets the data loaders for the training and validation data.

        Args:
            train_dataloader (DataLoader): The DataLoader object for the
                training data.
            val_dataloader (DataLoader): The DataLoader object for the
                validation data.
        """
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            tuple: The shape of an example feature array.
                derives shape from the GANDLF config's patch_size if available
        """

        # If we have a train dataloader with a dataset that has a gandlf_params attribute
        # with patch_size
        if (
            self.train_dataloader is not None
            and hasattr(self.train_dataloader, "dataset")
            and hasattr(self.train_dataloader.dataset, "gandlf_params")
            and "patch_size" in self.train_dataloader.dataset.gandlf_params
        ):
            # Return the patch size from GANDLF config
            return self.train_dataloader.dataset.gandlf_params["patch_size"]

        raise NotImplementedError(
            "Ensure gandlf params have to be defined in gandlf config file"
            " or implement get_feature_shape method in the GaNDLFDataLoaderWrapper class."
        )

    def get_train_loader(self, batch_size=None, num_batches=None):
        """Returns the data loader for the training data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).
            num_batches (int, optional): The number of batches for the data
                loader (default is None).

        Returns:
            DataLoader: The DataLoader object for the training data.
        """
        return self.train_dataloader

    def get_valid_loader(self, batch_size=None):
        """Returns the data loader for the validation data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).

        Returns:
            DataLoader: The DataLoader object for the validation data.
        """
        return self.val_dataloader

    def get_train_data_size(self):
        """Returns the total number of training samples.

        Returns:
            int: The total number of training samples or 0 if not loaded.
        """
        if self.train_dataloader is None or not hasattr(self.train_dataloader, "dataset"):
            return 0
        return len(self.train_dataloader.dataset)

    def get_valid_data_size(self):
        """Returns the total number of validation samples.

        Returns:
            int: The total number of validation samples or 0 if not loaded.
        """
        if self.val_dataloader is None or not hasattr(self.val_dataloader, "dataset"):
            return 0
        return len(self.val_dataloader.dataset)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.runtime module Runtime class."""

from typing import Callable, List

from openfl.experimental.workflow.interface.fl_spec import FLSpec
from openfl.experimental.workflow.interface.participants import Aggregator, Collaborator


class Runtime:
    def __init__(self):
        """Initializes the Runtime object.

        This serves as a base interface for runtimes that can run FLSpec flows.
        """
        pass

    @property
    def aggregator(self):
        """Returns the name of the aggregator.

        Raises:
            NotImplementedError: If the method is not implemented in a
                subclass.
        """
        raise NotImplementedError

    @aggregator.setter
    def aggregator(self, aggregator: Aggregator):
        """Sets the aggregator of the Runtime.

        Args:
            aggregator (Aggregator): The aggregator to be set.

        Raises:
            NotImplementedError: If the method is not implemented in a
                subclass.
        """
        raise NotImplementedError

    @property
    def collaborators(self):
        """Return the names of the collaborators. Don't give direct access to
        private attributes.

        Raises:
            NotImplementedError: If the method is not implemented in a
                subclass.
        """
        raise NotImplementedError

    @collaborators.setter
    def collaborators(self, collaborators: List[Collaborator]):
        """Sets the collaborators of the Runtime.

        Args:
            collaborators (List[Collaborator]): The collaborators to be set.

        Raises:
            NotImplementedError: If the method is not implemented in a
                subclass.
        """
        raise NotImplementedError

    def execute_task(
        self,
        flspec_obj: FLSpec,
        f: Callable,
        parent_func: Callable,
        instance_snapshot: List[FLSpec] = [],
        **kwargs,
    ):
        """Performs the execution of a task as defined by the implementation
        and underlying backend (single_process, ray, etc).

        Args:
            flspec_obj (FLSpec): Reference to the FLSpec (flow) object.
                Contains information about task sequence, flow attributes,
                that are needed to execute a future task.
            f (Callable): The next task to be executed within the flow.
            parent_func (Callable): The prior task executed in the flow.
            instance_snapshot (List[FLSpec], optional): A prior FLSpec state
                that needs to be restored from (i.e. restoring aggregator
                state after collaborator execution).
            **kwargs: Additional keyword arguments.

        Raises:
            NotImplementedError: If the method is not implemented in a
                subclass.
        """
        raise NotImplementedError



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""OpenFL Component Module."""

from importlib import util

from openfl.component.aggregator.aggregator import Aggregator
from openfl.component.aggregator.straggler_handling import (
    CutoffTimePolicy,
    PercentagePolicy,
    StragglerPolicy,
    WaitForAllPolicy,
)
from openfl.component.assigner.assigner import Assigner
from openfl.component.assigner.random_grouped_assigner import RandomGroupedAssigner
from openfl.component.assigner.static_grouped_assigner import StaticGroupedAssigner
from openfl.component.collaborator.collaborator import Collaborator

if util.find_spec("flwr") is not None:
    from openfl.component.connector.connector_flower import ConnectorFlower



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import tqdm

from openfl.federated import PyTorchTaskRunner
from openfl.utilities import TensorKey


def cross_entropy(output, target):
    """Binary cross-entropy metric.

    Args:
        output: The mode prediction
        target: The target (ground truth label)

    Returns:
        Binary cross-entropy with logits

    """
    return F.cross_entropy(input=output, target=target)


class PyTorchCNN(PyTorchTaskRunner):
    """Simple CNN for classification."""

    def __init__(self, device='cpu', **kwargs):
        """Initialize.

        Args:
            data: The data loader class
            device: The hardware device to use for training (Default = "cpu")
            **kwargs: Additional arguments to pass to the function

        """
        super().__init__(device=device, **kwargs)

        self.num_classes = self.data_loader.get_num_classes()
        self.init_network(device=self.device, **kwargs)
        self._init_optimizer()
        self.loss_fn = cross_entropy
        self.initialize_tensorkeys_for_functions()

    def _init_optimizer(self):
        """Initialize the optimizer."""
        self.optimizer = optim.Adam(self.parameters(), lr=1e-4)

    def init_network(self,
                     device,
                     print_model=True,
                     pool_sqrkernel_size=2,
                     conv_sqrkernel_size=5,
                     conv1_channels_out=20,
                     conv2_channels_out=50,
                     fc2_insize=500,
                     **kwargs):
        """Create the network (model).

        Args:
            device: The hardware device to use for training
            print_model (bool): Print the model topology (Default=True)
            pool_sqrkernel_size (int): Max pooling kernel size (Default=2),
                                       assumes square 2x2
            conv_sqrkernel_size (int): Convolutional filter size (Default=5),
                                       assumes square 5x5
            conv1_channels_out (int): Number of filters in first
                                      convolutional layer (Default=20)
            conv2_channels_out: Number of filters in second convolutional
                                layer (Default=50)
            fc2_insize (int): Number of neurons in the
                              fully-connected layer (Default = 500)
            **kwargs: Additional arguments to pass to the function

        FIXME: We are tracking only side lengths (rather than
        length and width) as we are assuming square
        shapes for feature and kernels.
        In order that all of the input and activation components are
        used (not cut off), we rely on a criterion: appropriate integers
        are divisible so that all casting to int performed below does no
        rounding (i.e. all int casting simply converts a float with '0'
        in the decimal part to an int.)

        (Note this criterion held for the original input sizes considered
        for this model: 28x28 and 32x32 when used with the default values
        above)
        """
        self.pool_sqrkernel_size = pool_sqrkernel_size
        channel = self.data_loader.get_feature_shape()[0]  # (channel, dim1, dim2)
        self.conv1 = nn.Conv2d(channel, conv1_channels_out, conv_sqrkernel_size, 1)

        # perform some calculations to track the size of the single channel activations
        # channels are first for pytorch
        conv1_sqrsize_in = self.feature_shape[-1]
        conv1_sqrsize_out = conv1_sqrsize_in - (conv_sqrkernel_size - 1)
        # a pool operation happens after conv1 out
        # (note dependence on 'forward' function below)
        conv2_sqrsize_in = int(conv1_sqrsize_out / pool_sqrkernel_size)

        self.conv2 = nn.Conv2d(conv1_channels_out, conv2_channels_out, conv_sqrkernel_size, 1)

        # more tracking of single channel activation size
        conv2_sqrsize_out = conv2_sqrsize_in - (conv_sqrkernel_size - 1)
        # a pool operation happens after conv2 out
        # (note dependence on 'forward' function below)
        l0 = int(conv2_sqrsize_out / pool_sqrkernel_size)
        self.fc1_insize = l0 * l0 * conv2_channels_out
        self.fc1 = nn.Linear(self.fc1_insize, fc2_insize)
        self.fc2 = nn.Linear(fc2_insize, self.num_classes)
        if print_model:
            print(self)
        self.to(device)

    def forward(self, x):
        """Forward pass of the model.

        Args:
            x: Data input to the model for the forward pass
        """
        x = F.relu(self.conv1(x))
        pl = self.pool_sqrkernel_size
        x = F.max_pool2d(x, pl, pl)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, pl, pl)
        x = x.view(-1, self.fc1_insize)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x

    def validate(self, col_name, round_num, input_tensor_dict, use_tqdm=False, **kwargs):
        """Validate.

        Run validation of the model on the local data.

        Args:
            col_name:            Name of the collaborator
            round_num:           What round is it
            input_tensor_dict:   Required input tensors (for model)
            use_tqdm (bool):     Use tqdm to print a progress bar (Default=True)

        Returns:
            global_output_dict:  Tensors to send back to the aggregator
            local_output_dict:   Tensors to maintain in the local TensorDB

        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        self.eval()
        val_score = 0
        total_samples = 0

        loader = self.data_loader.get_valid_loader()
        if use_tqdm:
            loader = tqdm.tqdm(loader, desc='validate')

        with torch.no_grad():
            for data, target in loader:
                samples = target.shape[0]
                total_samples += samples
                data, target = torch.tensor(data).to(
                    self.device), torch.tensor(target).to(
                    self.device, dtype=torch.int64)
                output = self(data)
                # get the index of the max log-probability
                pred = output.argmax(dim=1)
                val_score += pred.eq(target).sum().cpu().numpy()

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric', suffix)
        # TODO figure out a better way to pass
        #  in metric for this pytorch validate function
        output_tensor_dict = {
            TensorKey('acc', origin, round_num, True, tags):
                np.array(val_score / total_samples)
        }

        # Empty list represents metrics that should only be stored locally
        return output_tensor_dict, {}

    def reset_opt_vars(self):
        """Reset optimizer variables.

        Resets the optimizer state variables.

        """
        self._init_optimizer()

    def save_native(self, filepath):
        """
        Save model in a picked file specified by the filepath.
        Uses torch.save().

        Args:
            filepath (string)                 : Path to pickle file to be
                                                created by pt.save().
        Returns:
            None
        """
        torch.save(self, filepath)



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import hashlib
import os
from importlib import util
from typing import Generator
from urllib.parse import urlparse

from openfl.federated.data.sources.data_source import DataSource, DataSourceType


class S3DataSource(DataSource):
    """Class for S3 data"""

    def __init__(
        self,
        name,
        uri: str,
        endpoint=None,
        access_key_env_name=None,
        secret_key_env_name=None,
        secret_name=None,
        hash_func=None,
    ):
        if util.find_spec("boto3") is None:
            raise Exception(
                "'boto3' not installed.This package is necessary for interacting with AWS services."
            )
        import boto3

        super().__init__(DataSourceType.S3, name)
        self.uri = uri
        self.endpoint = endpoint
        self.access_key_env_name = access_key_env_name
        self.secret_key_env_name = secret_key_env_name
        self.secret_name = secret_name
        self.hash_func = hash_func

        access_key_id = os.environ.get(access_key_env_name) if access_key_env_name else None
        secret_access_key = os.environ.get(secret_key_env_name) if secret_key_env_name else None

        self._s3_client = boto3.client(
            "s3",
            endpoint_url=self.endpoint,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
        )

    def enumerate_files(self) -> Generator[str, None, None]:
        """Enumerate all files in the data source"""
        parsed = urlparse(self.uri)
        bucket_name = parsed.netloc
        prefix = parsed.path.lstrip("/")  # Remove leading slash

        paginator = self._s3_client.get_paginator("list_objects_v2")
        for page in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
            if "Contents" in page:
                for obj in page["Contents"]:
                    obj_key = obj["Key"]
                    if obj_key.endswith("/"):  # Ignore directories
                        continue
                    yield f"s3://{bucket_name}/{obj_key}"

    def _get_s3_etag(self, obj_path: str):
        parsed = urlparse(obj_path)
        bucket_name = parsed.netloc
        object_key = parsed.path.lstrip("/")
        response = self._s3_client.head_object(Bucket=bucket_name, Key=object_key)
        return response["ETag"]

    def _read_s3_object(self, obj_path: str):
        parsed = urlparse(obj_path)
        bucket_name = parsed.netloc
        object_key = parsed.path.lstrip("/")
        response = self._s3_client.get_object(Bucket=bucket_name, Key=object_key)
        return response["Body"].read()

    def compute_file_hash(self, path: str) -> str:
        if not self.hash_func:
            return self._get_s3_etag(path)
        else:
            data = self._read_s3_object(path)
            return self.hash_func(data).hexdigest()

    def read_blob(self, path):
        return self._read_s3_object(path)

    @classmethod
    def from_dict(cls, ds_dict: dict):
        hash_func_name = ds_dict.get("hash_func", None)
        if hash_func_name:
            hash_func = getattr(hashlib, hash_func_name, None)
        else:
            hash_func = None
        return cls(
            name=ds_dict["name"],
            uri=ds_dict["uri"],
            endpoint=ds_dict.get("endpoint", None),
            access_key_env_name=ds_dict.get("access_key_env_name", None),
            secret_key_env_name=ds_dict.get("secret_key_env_name", None),
            hash_func=hash_func,
        )



# Copyright (C) 2022-2024 TU Darmstadt
# SPDX-License-Identifier: Apache-2.0

# -----------------------------------------------------------
# Primary author: Phillip Rieger <phillip.rieger@trust.tu-darmstadt.de>
# Co-authored-by: Torsten Krauss <torsten.krauss@uni-wuerzburg.de>
# ------------------------------------------------------------
from enum import Enum
import math
from copy import deepcopy
import numpy as np
import torch
from matplotlib import pyplot as plt
from scipy.stats import kstest, levene, ttest_ind
from sklearn import preprocessing
from sklearn.cluster import AgglomerativeClustering
from sklearn.decomposition import PCA
from torch import cosine_similarity


class DistanceMetric(str, Enum):
    """Enum to identify distance metrics necessary in this project"""
    COSINE = 'cosine'
    EUCLIDEAN = 'euclid'


class DistanceHandler:
    """Helper, that calculates distances between two tensors."""

    @staticmethod
    def __get_euclid_distance(t1: torch.Tensor, t2: torch.Tensor) -> float:
        t = t1.view(-1) - t2.view(-1)
        return torch.norm(t, 2).cpu().item()

    @staticmethod
    def __get_cosine_distance(t1: torch.Tensor, t2: torch.Tensor) -> float:
        t1 = t1.view(-1).reshape(1, -1)
        t2 = t2.view(-1).reshape(1, -1)
        return 1 - cosine_similarity(t1, t2).cpu().item()

    @staticmethod
    def get_distance(distance, t1: torch.Tensor, t2: torch.Tensor) -> float:
        """Factory Method for Distances"""
        if distance == DistanceMetric.COSINE:
            return DistanceHandler.__get_cosine_distance(t1, t2)
        if distance == DistanceMetric.EUCLIDEAN:
            return DistanceHandler.__get_euclid_distance(t1, t2)

        raise Exception(f"Extractor for {distance} not implemented yet.")


class CrowdGuardClientValidation:

    @staticmethod
    def __distance_global_model_final_metric(distance_type: str, prediction_matrix,
                                             prediction_global_model, sample_indices_by_label,
                                             own_index):
        """
        Calculates the distance matrix containing the metric for CrowdGuard
        with dimensions label x model x layer x values
        """

        sample_count = len(prediction_matrix)
        model_count = len(prediction_matrix[0])
        layer_count = len(prediction_matrix[0][0])

        # We create a distance matrix with distances between global and local models
        # of the dimensions sample x model x layer x values
        global_distance_matrix = [[[0.] * layer_count for _ in range(model_count)]
                                  for _ in range(sample_count)]
        # 1. calculate distances between predictions of global model and each local model
        for s_i, s in enumerate(prediction_matrix):
            g = prediction_global_model[s_i]
            for m_i, m in enumerate(s):
                for l_i, l in enumerate(m):
                    distance = DistanceHandler.get_distance(distance_type, l, g[
                        l_i])  # either euclidean or cosine distance
                    global_distance_matrix[s_i][m_i][l_i] = distance  # line 18

        # 2. Sort the sample-wise distances by the label of the sample
        for label, sample_list in sample_indices_by_label.items():
            # First pick the samples from the global predictions
            global_distance_matrix_for_label_helper = [
                [[0.] * len(sample_list) for _ in range(layer_count)] for _ in
                range(model_count)]

            s_i_new = 0
            for s_i, s in enumerate(global_distance_matrix):
                if s_i not in sample_list:
                    continue
                for m_i, mi in enumerate(s):
                    for l_i, l in enumerate(mi):
                        global_distance_matrix_for_label_helper[m_i][l_i][s_i_new] = l
                s_i_new += 1

        # We produce the first relative matrix
        sample_relation_matrix = [[[0.] * layer_count for _ in range(model_count)] for _ in
                                  range(sample_count)]

        # 3. divide by distances of this client to use its values as reference
        for s_i, s in enumerate(global_distance_matrix):
            distances_for_own_models_predictions = s[own_index]
            for m_j, mj in enumerate(s):
                for l_i, l in enumerate(mj):
                    relation = 0
                    if distances_for_own_models_predictions[l_i] != 0:
                        relation = l / distances_for_own_models_predictions[l_i]
                    sample_relation_matrix[s_i][m_j][l_i] = relation  # line 21

        # We produce the Label average
        # We produce a matrix with not all samples, but mean all the samples, so that we have a
        # Matrix per label
        sample_relation_matrix_for_label = {}

        # 4. Transpose matrix as preparation for averaging
        for label, sample_list in sample_indices_by_label.items():
            sample_relation_matrix_for_label[label] = [[0.] * layer_count for _ in
                                                       range(model_count)]
            sample_relation_matrix_for_label_helper = [
                [[0.] * len(sample_list) for _ in range(layer_count)] for _ in range(model_count)]
            # transpose dimensions of distance matrix, before we had (sample,model, layer) and
            # we transpose it to (model,layer,sample)
            s_i_new = 0
            for s_i, s in enumerate(sample_relation_matrix):
                if s_i not in sample_list:
                    continue
                for m_j, mj in enumerate(s):
                    for l_i, l in enumerate(mj):
                        sample_relation_matrix_for_label_helper[m_j][l_i][s_i_new] = l
                s_i_new += 1

            # 5. Average over all samples from the same label (basically kick-out the last
            # dimension)
            for m_j, mj in enumerate(sample_relation_matrix_for_label_helper):
                for l_i, l in enumerate(mj):
                    sample_relation_matrix_for_label[label][m_j][l_i] = np.mean(l).item()

        avg_sample_relation_matrix_squared_negative_models_first = {}

        # 6. subtract 1 (mainly for cosine distances) and square (but keep the sign)
        for label, label_values in sample_relation_matrix_for_label.items():
            avg_sample_relation_matrix_squared_negative_models_first[label] = [[0.] * layer_count
                                                                               for _ in
                                                                               range(model_count)]
            for m_j, mj in enumerate(label_values):
                for l_i, l in enumerate(mj):
                    x = l - 1
                    relation = x * x
                    relation = relation if x >= 0 else relation * (-1)
                    avg_sample_relation_matrix_squared_negative_models_first[label][m_j][
                        l_i] = relation
        return avg_sample_relation_matrix_squared_negative_models_first

    @staticmethod
    def __predict_for_single_model(model, local_data, device):
        """
        Returns
        - A matrix with Deep Layer Outputs with dimensions sample x layer x values.
        - The labels for all samples in the client's training dataset
        - The number of layers defined in the model
        """
        num_layers = None
        sample_label_list = []
        predictions = []
        model.eval()
        model = model.to(device)
        number_of_previous_samples = 0
        for batch_id, batch in enumerate(local_data):
            data, target = batch
            data, target = data.to(device), target.to(device)
            output = model.predict_internal_states(data)
            if num_layers is None:
                num_layers = len(output)
            assert num_layers == len(output)

            for layer_output_index, layer_output_values in enumerate(output):
                for idx in range(target.shape[0]):
                    sample_idx = number_of_previous_samples + idx
                    assert len(predictions) >= sample_idx
                    if len(predictions) == sample_idx:
                        assert layer_output_index == 0
                        predictions.append([])

                    if layer_output_index == 0:
                        expected_predictions = sample_idx + 1
                    else:
                        expected_predictions = number_of_previous_samples + target.shape[0]
                    assert_msg = f'{len(predictions)} vs. {sample_idx} ({idx} {batch_id} '
                    assert_msg += f'{layer_output_index} {number_of_previous_samples})'
                    assert len(predictions) == expected_predictions, assert_msg
                    assert_msg = f'{len(predictions[sample_idx])} {layer_output_index} '
                    assert_msg += f'{sample_idx} {batch_id} {idx} {number_of_previous_samples}'
                    assert len(predictions[sample_idx]) == layer_output_index, assert_msg
                    value = layer_output_values[idx].clone().detach().cpu()
                    predictions[sample_idx].append(value)
            number_of_previous_samples += target.shape[0]
            for t in target:
                sample_label_list.append(t.detach().clone().cpu().item())
        model.cpu()
        return predictions, sample_label_list, num_layers

    @staticmethod
    def __do_predictions(models, global_model, local_data, device):
        """
        Returns
        - The Deep Layer Outputs for all models in a matrix of dimension
          sample x model x layer x value
        - The Deep Layer Outputs of the global model int he dimension sample x layer x value
        - A dict containing lists of sample indices for each label class
        - The number of layers from the model
        """
        all_models_predictions = []
        for model_index, model in enumerate(models):
            predictions, _, _ = CrowdGuardClientValidation.__predict_for_single_model(model,
                                                                                      local_data,
                                                                                      device)
            for sample_index, layer_predictions_for_sample in enumerate(predictions):
                if sample_index >= len(all_models_predictions):
                    assert model_index == 0
                    assert len(all_models_predictions) == sample_index
                    all_models_predictions.append([])
                all_models_predictions[sample_index].append(layer_predictions_for_sample)
        tmp = CrowdGuardClientValidation.__predict_for_single_model(global_model, local_data,
                                                                    device)
        global_model_predictions, sample_label_list, n_layers = tmp
        sample_indices_by_label = {}
        for s_i, label in enumerate(sample_label_list):
            if label not in sample_indices_by_label.keys():
                sample_indices_by_label[label] = []
            sample_indices_by_label[label].append(s_i)

        return all_models_predictions, global_model_predictions, sample_indices_by_label, n_layers

    @staticmethod
    def __prune_poisoned_models(num_layers, total_number_of_clients, own_client_index,
                                distances_by_metric, verbose=False):
        detected_poisoned_models = []
        for distance_type in distances_by_metric.keys():

            # First load the distance Matrix for this client and the samples by labels.
            distance_matrix_la_m_l = distances_by_metric[distance_type]

            # We put all of our labels into one big row.
            layer_length = num_layers * len(distance_matrix_la_m_l)
            dist_matrix_m_lcon = [[0.] * layer_length for _ in range(total_number_of_clients)]
            label_count = 0
            for label_x, dist_matrix_m_l_for_label in distance_matrix_la_m_l.items():
                for model_idx, model_values in enumerate(dist_matrix_m_l_for_label):
                    for layer_idx, layer in enumerate(model_values):
                        dist_matrix_m_lcon[model_idx][layer_idx + label_count * num_layers] = layer
                label_count = label_count + 1

            dist_matrix_m_l = dist_matrix_m_lcon

            client_indices = [x for x, value in enumerate(dist_matrix_m_l) if
                              x != own_client_index]
            pruned_indices = []
            has_malicious_model = True
            new_round_needed = True
            prune_idx = 0

            max_pruning_count = int(math.floor((len(dist_matrix_m_l) - 1) / 2))

            while has_malicious_model and new_round_needed:
                # unique
                pruned_indices_local = deepcopy(pruned_indices)
                # Ignore the own label again and the pruned indices
                pruned_cluster_input_m_l = [value for x, value in
                                            enumerate(dist_matrix_m_l) if
                                            x != own_client_index and x not in pruned_indices]
                pruned_client_indices = [x for x, value in
                                         enumerate(dist_matrix_m_l) if
                                         x != own_client_index and x not in pruned_indices]

                if len(pruned_cluster_input_m_l) <= 1:
                    break

                layer_values = {}

                for m in pruned_cluster_input_m_l:
                    for l_i, l in enumerate(m):
                        if l_i not in layer_values.keys():
                            layer_values[l_i] = []
                        layer_values[l_i].append(l)

                median_layer_values = []

                for l_i, l_values in layer_values.items():
                    median_layer_values.append(np.median(l_values).item())

                median_graph = list(median_layer_values)

                pca_list = []
                for m in pruned_cluster_input_m_l:
                    pca_list.append(m)

                pca_list.append(median_graph)

                scaled_data = preprocessing.scale(pca_list)

                pca = PCA()
                pca.fit(scaled_data)
                pca_data = pca.transform(scaled_data)

                cluster_input = []
                cluster_input_plain = []
                pca_one_data = pca_data.T[0]
                for pca_one_value in pca_one_data:
                    cluster_input.append([pca_one_value])
                    cluster_input_plain.append(pca_one_value)

                # Significance tests
                median_val = np.median(cluster_input_plain)
                if verbose:
                    print(f'cluster_input_plain={cluster_input_plain}')
                x_values = []
                y_values = []
                for value in cluster_input_plain:
                    # Split the samples into two groups
                    distance_value = abs(value - median_val)
                    if value >= median_val:
                        x_values.append(distance_value)
                    else:
                        y_values.append(distance_value)
                print(f'Distance: {distance_type}, use y {len(y_values)}: {y_values}')
                print(f'Distance: {distance_type}, use x {len(x_values)}: {x_values}')

                # Statistical tests
                t_value, t_p_value = ttest_ind(x_values, y_values)
                ks_value, ks_p_value = kstest(x_values, y_values)
                barlett_value, bartlett_p_value = levene(x_values, y_values)

                # Outlier tests
                # Creating boxplot
                bp_result = plt.boxplot(cluster_input_plain, whis=5.5)
                fliers = bp_result['fliers'][0].get_ydata()
                outlier_boxplot = len(fliers)
                plt.close()

                # Outlier based on variance
                deviation_mean = np.mean(cluster_input_plain)
                deviation_std = abs(np.std(cluster_input_plain))

                max_dist_rule_factor = 0
                for cip in cluster_input_plain:
                    cip_abs = abs(cip - deviation_mean)
                    rule_factor = cip_abs / deviation_std
                    if max_dist_rule_factor < rule_factor:
                        max_dist_rule_factor = rule_factor

                outlier_three_sigma = max_dist_rule_factor

                has_malicious_model_t_threshold = False
                if t_p_value < 0.01:
                    has_malicious_model_t_threshold = True
                has_malicious_model_ks_threshold = False
                if ks_p_value < 0.01:
                    has_malicious_model_ks_threshold = True
                has_malicious_model_bartlett_threshold = False
                if bartlett_p_value < 0.01:
                    has_malicious_model_bartlett_threshold = True

                has_boxplot_outlier = False
                has_three_sigma_outlier = False

                if outlier_boxplot > 0:
                    has_boxplot_outlier = True
                if outlier_three_sigma >= 3:
                    has_three_sigma_outlier = True

                # Choose exit criterium
                has_malicious_model = (has_malicious_model_t_threshold
                                       or has_malicious_model_ks_threshold
                                       or has_malicious_model_bartlett_threshold
                                       or has_boxplot_outlier
                                       or has_three_sigma_outlier)

                ac_e = AgglomerativeClustering(n_clusters=2, distance_threshold=None,
                                               compute_full_tree=True,
                                               metric="euclidean", memory=None,
                                               connectivity=None,
                                               linkage='single',
                                               compute_distances=True).fit(cluster_input)
                ac_e_labels: list = ac_e.labels_.tolist()
                median_value_cluster_label = ac_e_labels[-1]
                ac_e_malicious_class_indices = [idx for idx, val in enumerate(ac_e_labels) if
                                                val != median_value_cluster_label]

                for m_j, value in enumerate(pruned_client_indices):
                    if m_j in ac_e_malicious_class_indices:
                        pruned_indices_local.append(value)

                pruned_indices_local = list(set(pruned_indices_local))

                # If we now prune more than half, we stop and remove the best items from the last
                # pruning list.
                pruned_too_much = True
                if len(pruned_indices_local) > max_pruning_count:
                    dist_values_of_pruned_models = []
                    for midx in ac_e_malicious_class_indices:
                        dist_to_median = abs(cluster_input[midx][0] - cluster_input[-1][0])
                        dist_values_of_pruned_models.append(dist_to_median)

                    sorted_dist_values_of_pruned_models = list(dist_values_of_pruned_models)
                    sorted_dist_values_of_pruned_models.sort()

                    sorted_ac_e_malicious_class_indices = []
                    for sdv in sorted_dist_values_of_pruned_models:
                        dvidx = dist_values_of_pruned_models.index(sdv)
                        for m_j, value in enumerate(pruned_client_indices):
                            if m_j == ac_e_malicious_class_indices[dvidx]:
                                sorted_ac_e_malicious_class_indices.append(value)
                    overflowed_count = len(pruned_indices_local) - max_pruning_count
                    for oc in range(overflowed_count):
                        # Get the values of the clusters and remove the nearest ones
                        # from pruned_indices_local
                        pruned_indices_local.remove(sorted_ac_e_malicious_class_indices[-1])
                        del sorted_ac_e_malicious_class_indices[-1]
                    pruned_too_much = False

                still_pruning = len(pruned_indices) < len(pruned_indices_local)
                new_round_needed = still_pruning and pruned_too_much
                if has_malicious_model and new_round_needed:
                    pruned_indices = pruned_indices_local

                prune_idx += 1

            # Analyze the voting
            for _, value in enumerate(client_indices):
                if value in pruned_indices:
                    detected_poisoned_models.append(value)

        return list(set(detected_poisoned_models))

    @staticmethod
    def validate_models(global_model, models, own_client_index, local_data, device):
        tmp = CrowdGuardClientValidation.__do_predictions(models, global_model, local_data, device)
        prediction_matrix, global_model_predictions, sample_indices_by_label, num_layers = tmp
        distances_by_metric = {}
        for dist_type in [DistanceMetric.COSINE, DistanceMetric.EUCLIDEAN]:
            calculated_distances = CrowdGuardClientValidation.__distance_global_model_final_metric(
                dist_type,
                prediction_matrix,
                global_model_predictions,
                sample_indices_by_label,
                own_client_index)
            distances_by_metric[dist_type] = calculated_distances
        result = CrowdGuardClientValidation.__prune_poisoned_models(num_layers, len(models),
                                                                    own_client_index,
                                                                    distances_by_metric)
        return result



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow as tf
import keras

from openfl.federated import KerasTaskRunner

import logging
logger = logging.getLogger(__name__)

class Keras2DUNet(KerasTaskRunner):
    """Initialize.

    Args:
        **kwargs: Additional parameters to pass to the function

    """

    def __init__(self, **kwargs):
        """Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function

        """
        super().__init__(**kwargs)

        self.model = self.build_model(self.data_loader.get_feature_shape(), use_upsampling=True, **kwargs)
        self.model.summary(print_fn=logger.info, line_length=120)
        self.initialize_tensorkeys_for_functions()


    def build_model(self, input_shape,
                use_upsampling=False,
                n_cl_out=1,
                dropout=0.2,
                activation_function='relu',
                seed=0xFEEDFACE,
                depth=5,
                dropout_at=None,
                initial_filters=32,
                batch_norm=True,
                **kwargs):
        """Define the TensorFlow model.

        Args:
            input_shape: input shape of the model
            use_upsampling (bool): True = use bilinear interpolation;
                                False = use transposed convolution (Default=False)
            n_cl_out (int): Number of channels in input layer (Default=1)
            dropout (float): Dropout percentage (Default=0.2)(Default = True)
            activation_function: The activation function to use after convolutional layers (Default='relu')
            seed: random seed (Default=0xFEEDFACE)
            depth (int): Number of max pooling layers in encoder (Default=5)
            dropout_at: Layers to perform dropout after (Default=[2,3])
            initial_filters (int): Number of filters in first convolutional layer (Default=32)
            batch_norm (bool): True = use batch normalization (Default=True)
            **kwargs: Additional parameters to pass to the function

        """
        if dropout_at is None:
            dropout_at = [2, 3]

        inputs = keras.layers.Input(shape=input_shape, name='Images')

        if activation_function == 'relu':
            activation = tf.nn.relu
        elif activation_function == 'leakyrelu':
            activation = tf.nn.leaky_relu

        params = {
            'activation': activation,
            'kernel_initializer': keras.initializers.he_uniform(seed=seed),
            'kernel_size': (3, 3),
            'padding': 'same',
        }

        convb_layers = {}

        net = inputs
        filters = initial_filters
        for i in range(depth):
            name = f'conv{i + 1}a'
            net = keras.layers.Conv2D(name=name, filters=filters, **params)(net)
            if i in dropout_at:
                net = keras.layers.Dropout(dropout)(net)
            name = f'conv{i + 1}b'
            net = keras.layers.Conv2D(name=name, filters=filters, **params)(net)
            if batch_norm:
                net = keras.layers.BatchNormalization()(net)
            convb_layers[name] = net
            # only pool if not last level
            if i != depth - 1:
                name = f'pool{i + 1}'
                net = keras.layers.MaxPooling2D(name=name, pool_size=(2, 2))(net)
                filters *= 2

        # do the up levels
        filters //= 2
        for i in range(depth - 1):
            if use_upsampling:
                up = keras.layers.UpSampling2D(
                    name=f'up{depth + i + 1}', size=(2, 2))(net)
            else:
                up = keras.layers.Conv2DTranspose(
                    name='transConv6', filters=filters,
                    kernel_size=(2, 2), strides=(2, 2), padding='same')(net)
            net = keras.layers.concatenate(
                [up, convb_layers[f'conv{depth - i - 1}b']],
                axis=-1
            )
            net = keras.layers.Conv2D(
                name=f'conv{depth + i + 1}a',
                filters=filters, **params)(net)
            net = keras.layers.Conv2D(
                name=f'conv{depth + i + 1}b',
                filters=filters, **params)(net)
            filters //= 2
        net = keras.layers.Conv2D(name='Mask', filters=n_cl_out,
                                    kernel_size=(1, 1),
                                    activation='sigmoid')(net)
        model = keras.models.Model(inputs=[inputs], outputs=[net])


        self.optimizer = keras.optimizers.RMSprop(1e-2)
        model.compile(
            loss=self.dice_coef_loss,
            optimizer=self.optimizer,
            metrics=["acc"]
        )

        return model

    def dice_coef_loss(self, y_true, y_pred, smooth=1.0):
        """Dice coefficient loss.

        Calculate the -log(Dice Coefficient) loss

        Args:
            y_true: Ground truth annotation array
            y_pred: Prediction array from model
            smooth (float): Laplace smoothing factor (Default=1.0)
        Returns:
            float: -log(Dice coefficient) metric
        """
        intersection = tf.reduce_sum(y_true * y_pred, axis=(1, 2, 3))

        term1 = -tf.math.log(tf.constant(2.0) * intersection + smooth)
        term2 = tf.math.log(tf.reduce_sum(y_true, axis=(1, 2, 3))
                    + tf.reduce_sum(y_pred, axis=(1, 2, 3)) + smooth)

        term1 = tf.reduce_mean(term1)
        term2 = tf.reduce_mean(term2)

        loss = term1 + term2

        return loss



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""AggregatorClientInterface module."""

from abc import ABC, abstractmethod
from typing import Any, List, Tuple


class AggregatorClientInterface(ABC):
    @abstractmethod
    def ping(self):
        """
        Ping the aggregator to check connectivity.
        """
        pass

    @abstractmethod
    def get_tasks(self) -> Tuple[List[Any], int, int, bool]:
        """
        Retrieves tasks for the given collaborator client.
        Returns a tuple: (tasks, round_number, sleep_time, time_to_quit)
        """
        pass

    @abstractmethod
    def get_aggregated_tensor(
        self,
        tensor_name: str,
        round_number: int,
        report: bool,
        tags: List[str],
        require_lossless: bool,
    ) -> Any:
        """
        Retrieves the aggregated tensor.
        """
        pass

    @abstractmethod
    def send_local_task_results(
        self,
        round_number: int,
        task_name: str,
        data_size: int,
        named_tensors: List[Any],
    ) -> Any:
        """
        Sends local task results.
        Parameters:
          collaborator_name: Name of the collaborator.
          round_number: The current round.
          task_name: Name of the task.
          data_size: Size of the data.
          named_tensors: A list of tensors (or named tensor objects).
        Returns a SendLocalTaskResultsResponse.
        """
        pass

    @abstractmethod
    def send_message_to_server(self, openfl_message: Any, collaborator_name: str) -> Any:
        """
        Forwards a converted message from the local client to the OpenFL server and returns the
        response.
        Args:
            openfl_message: The converted message to be sent to the OpenFL server (InteropMessage
                proto).
            collaborator_name: The name of the collaborator.
        Returns:
            The response from the OpenFL server (InteropMessage proto).
        """
        pass
